/*****************************************************
Compiler            : CodeVisionAVR 2.04.6 Evaluation
Chip type           : ATmega8
Clock frequency     : 8 MHz (int. RC-OSC)
*****************************************************/

//***************************************************
// Spiel "Memo"
// auf Basis des Franzis-Pingpong
// http://www.elo-web.de/ping-pong-start
// 1 zustzlicher Taster erforderlich, 
// Anschluss zwischen PORTD.3 und GND
//***************************************************

#include <mega8.h>
#include <delay.h>
#include <sleep.h>
#include "memo.h"
#include "main.h"


#define P_CLK    PORTB.3
#define P_DATA   PORTB.4
#define P_STROBE PORTB.2

#define PIN_COIN     PIND.2
#define PULLUP_COIN  PORTD.2
#define DIR_COIN     DDRD.2
#define COIN_INSERTED (PIN_COIN == 0)

#define PIN_KEY     PIND.3
#define PULLUP_KEY  PORTD.3
#define KEY_PRESSED (PIN_KEY == 0)


unsigned int leds[WIDTH];       // current state of each LED (organized in columns)
unsigned int ledsBlink[WIDTH];  // current blink state of each LED (organized in columns)



flash unsigned char decimals[] =
{                   // Font for decimal values
0x1F, 0x11, 0x1F,   // 0
0x00, 0x00, 0x1F,   // 1
0x1D, 0x15, 0x17,   // 2
0x15, 0x15, 0x1F,   // 3
0x07, 0x04, 0x1F,   // 4
0x17, 0x15, 0x1D,   // 5
0x1F, 0x15, 0x1D,   // 6
0x01, 0x01, 0x1F,   // 7
0x1F, 0x15, 0x1F,   // 8
0x17, 0x15, 0x1F    // 9
};


// clear the whole screen, all LEDs off
void clearScreen(void)
{
   unsigned char i;
   
   for(i = 0; i < (WIDTH); i++)
   {
      leds[i] = 0;
   }
}


// wait until key is pressed
void waitUntilKeypressed(void)
{
   while(!KEY_PRESSED);
   delay_ms(20);
}


// wait until key is released
void waitUntilKeyreleased(void)
{
   while(KEY_PRESSED);
   delay_ms(20);
}


// check if key is pressed
// return: 0=key not pressed, 1=key pressed
unsigned char keyPressed(void)
{
   unsigned char result = 0;
   
   if(KEY_PRESSED)
      result = 1;
   delay_ms(20);
   
   return result;
}


// Timer 2 overflow interrupt service routine
// multiplexing of LEDs
// approx. every 1ms
interrupt [TIM2_OVF] void timer2_ovf_isr(void)
{
   static unsigned char count = 0;
   unsigned char cValue, cValue2;
   static unsigned char cBlink = 0x00;
   static unsigned char cBlinkcnt = 0;

   cBlinkcnt++;
   if(cBlinkcnt >= 250)
   {
      cBlinkcnt = 0;
      cBlink = ~cBlink;
   }

   count++;
   if(count >= 12)
   {
      count = 0;
      P_DATA = 0;
   }
   else
   {
      P_DATA = 1;
   }
   P_CLK = 1;
   P_CLK = 0;
   PORTC &= (~0x0F);
   PORTD &= (~0xF0);
   PORTB &= (~0x03);
   P_STROBE = 1;
   P_STROBE = 0;
   cValue = (unsigned char)leds[count] & 0x0F;
   cValue2 = (unsigned char)ledsBlink[count] & cBlink;
   cValue &= (~cValue2);
   PORTC |= cValue;
   cValue = (unsigned char)leds[count] & 0xF0;
   cValue &= (~cValue2);
   PORTD |= cValue;
   cValue = (unsigned char)(leds[count]>>8) & 0x03;
   cValue2 = (unsigned char)(ledsBlink[count]>>8) & cBlink;
   cValue &= (~cValue2);
   PORTB |= cValue;
}


// initialization of Timer2
void initTimer2(void)
{
// Timer/Counter 2 initialization
// Clock source: System Clock
// Clock value: 250,000 kHz
// Mode: Normal top=FFh
// OC2 output: Disconnected
ASSR=0x00;
TCCR2=0x03;
TCNT2=0x00;
OCR2=0x00;

// Timer(s)/Counter(s) Interrupt(s) initialization
TIMSK=0x40;
}



// set controller in Powerdown mode
void setStandbyMode(void)
{
   TCCR2 &= 0xF8;   //stop Timer2
   PORTC = 0;
   PORTD = 0;
   PORTB = 0;
   ADCSRA &= 0x7F;  // stopADC
   DDRD.2 = 0;
   PORTD.2 = 1;
   GICR |= 0x40;    //enable INT0
   sleep_enable();
   powerdown();
   GICR &= (~0x40);   //disable INT0
   PULLUP_KEY = 1;
   DDRD.2 = 1;
   PORTD.2 = 0;
   ADCSRA |= 0x80;  //startADC
   TCCR2 |= 0x03;   //start Timer2
}


// External Interrupt 0 service routine
interrupt [EXT_INT0] void ext_int0_isr(void)
{
}


// initialization of INT0
void initInt0(void)
{
   // External Interrupt(s) initialization
   // INT0: disabled
   // INT0 Mode: Low level
   GICR &= (~0x40);   //disable INT0
   MCUCR &= 0xFC;
   GIFR=0x40;
}


// initialization of ports
void init(void)
{
PORTB = 0;
DDRB = 0x1F;
PORTC = 0;
DDRC = 0x0F;
PORTD = 0;
DDRD = 0xF4;
PULLUP_KEY = 1;
// Analog Comparator initialization
// Analog Comparator: Off
// Analog Comparator Input Capture by Timer/Counter 1: Off
ACSR=0x80;
SFIOR=0x00;
}



#define ADC_VREF_TYPE 0x00

// initialization of ADC
void initADC(void)
{
   // ADC initialization
   // ADC Clock frequency: 125,000 kHz
   // ADC Voltage Reference: AREF pin
   ADMUX=ADC_VREF_TYPE & 0xff;
   ADCSRA=0x86;
}


// Read the AD conversion result
// adc_input: ADC channel (0..7)
// return: ADC value (0..1023)
unsigned int read_adc(unsigned char adc_input)
{
   ADMUX=adc_input | (ADC_VREF_TYPE & 0xff);
   // Delay needed for the stabilization of the ADC input voltage
   delay_us(10);
   // Start the AD conversion
   ADCSRA|=0x40;
   // Wait for the AD conversion to complete
   while ((ADCSRA & 0x10)==0);
   ADCSRA|=0x10;
   
   return ADCW;
}


// select x-position of a card
// return: selected sensor index (0...5)
unsigned char selectX(void)
{
   unsigned char x;                        
                                           
   x = 5 - (read_adc(7) / (1023/6 + 1));

   return(x);
}


// select y-position of a card
// return: selected value (0..4)
unsigned char selectY(void)
{
   unsigned char y;                        
   
   y = (read_adc(6) / (1023/5 + 1));
   
   return(y);
}


// display a decimal value
// x: decimal value (0..19)
// row: row to begin displaying value
// col: column to begin displaying value
void showDecimal(unsigned char x, unsigned char row, unsigned char col)
{
   unsigned char j, value;
   
   if(x > 19)
      x = 19;
   
   if(x >= 10) 
   {      // show "1"
      leds[col++] = ((unsigned int)0x001F << row);
      leds[col++] = 0;
   }

   value = x % 10;
   for(j = 0; j < 3; j++)
   {      // show decimal using stored font
      leds[col++] = ((unsigned int)decimals[value*3 + j] << row);
   }          
}




// main routine
void main(void)
{
   
   init();
   initTimer2();
   initADC();
   initInt0();

   clearScreen();
   
 
   #asm("sei")      // Global enable interrupts


   while(1)
   {    
      playMemo();
      setStandbyMode();
   }   
}
