/*****************************************************
Compiler            : CodeVisionAVR 2.04.6 Evaluation
Chip type           : ATmega8
Clock frequency     : 8 MHz (int. RC-OSC)
*****************************************************/

//***************************************************
// Spiel "Memo"
// auf Basis des Franzis-Pingpong
// http://www.elo-web.de/ping-pong-start
// 1 zustzlicher Taster erforderlich, 
// Anschluss zwischen PORTD.3 und GND
//***************************************************

#include <mega8.h>
#include <stdlib.h>
#include <delay.h>
#include "memo.h"
#include "main.h"


#define GAME_NOT_STARTED 0
#define GAME_RUNNING     1
#define GAME_OVER        2

unsigned char gameStatus = GAME_NOT_STARTED;  // current game status

#define COLUMNS  6              // number of columns for cards
#define ROWS     5              // number of rows for cards
#define MAX_ROW (ROWS-1)        // max row for cards
#define MAX_COLUMN (COLUMNS-1)  // max column for cards

#define CARDSELECTED 0x01       // status which each card can have
#define CARDPRESENT  0x02
#define CARDHIDDEN   0x04

#define BLINKING 1

unsigned char cYpos = 0;
unsigned char cXpos = 0;
unsigned char cCurrentXpos = 0;
unsigned char cCurrentYpos = 0;
unsigned char cLastYpos = 0;
unsigned char cLastXpos = 0;
typedef struct
{
   flash unsigned char *image;
   unsigned char status;
} T_CARD;
T_CARD arCards[ROWS][COLUMNS];  // array containing status of each card

unsigned char cScorePlayer1 = 0;
unsigned char cScorePlayer2 = 0;
unsigned char cCurrentPlayer = 1;
unsigned char cRemainingPairs = ROWS*COLUMNS/2;

extern unsigned int leds[];       // current state of each LED (organized in columns)
extern unsigned int ledsBlink[];  // current blink state of each LED (organized in columns)


flash unsigned char arImages[] =  // card images
{
   0x00, 0x7F, 0x3E, 0x1C, 0x08,
   0x08, 0x1C, 0x3E, 0x7F, 0x00,
   0x1C, 0x3E, 0x3E, 0x3E, 0x1C,
   0x10, 0x38, 0x54, 0x10, 0x1F,
   0x04, 0x02, 0x7F, 0x02, 0x04,
   0x10, 0x20, 0x7F, 0x20, 0x10,
   0x08, 0x08, 0x2A, 0x1C, 0x08,
   0x08, 0x1C, 0x2A, 0x08, 0x08,
   0x14, 0x08, 0x3E, 0x08, 0x14,
   0x60, 0x60, 0x7F, 0x02, 0x0C,
   0x10, 0x1E, 0x3F, 0x1E, 0x10,
   0x5C, 0x62, 0x02, 0x62, 0x5C,
   0x1C, 0x3E, 0x7C, 0x3E, 0x1C,
   0x7C, 0x4A, 0x49, 0x4A, 0x7C,
   0x63, 0x14, 0x08, 0x14, 0x63
};


flash unsigned int logoMemo[WIDTH] =   // start logo
{
   0x01F, 0x3E2, 0x044, 0x082, 0x05f, 0x3E0, 0x01F, 0x1D5, 0x235, 0x235, 0x1C0, 0x000
};


flash unsigned int arrowLeftLogo[WIDTH] =  // logo for player selection left
{ 0x0010, 0x0038, 0x007C, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 };

flash unsigned int arrowRightLogo[WIDTH] = // logo for player selection right
{ 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x007C, 0x0038, 0x0010 };


// show a logo on screen
// pLogo: pointer to array with logo data
// state: 0 = static logo, BLINKING = blinking logo 
void showLogo(flash unsigned int *pLogo, unsigned char state)
{
   unsigned char i;

   for(i = 0; i < WIDTH; i++)
   {        
      leds[i] = pLogo[i];
      if(state == BLINKING)
         ledsBlink[i] = 0xFFFF;
      else
         ledsBlink[i] = 0;
   }
}



// show a field as selected (blinking ON)
// row: row of field
// col: column postion of field
void selectField(unsigned char row, unsigned char col)
{
   leds[col*2] |= ((unsigned int)0x0001 << (row*2));
   ledsBlink[col*2] |= ((unsigned int)0x0001 << (row*2));
}


// show a field as unselected (blinking OFF)
// row: row of field
// col: column postion of field
void unselectField(unsigned char row, unsigned char col)
{
   ledsBlink[col*2] &= (~((unsigned int)0x0001 << (row*2)));
}


// show an empty field (LED OFF)
// row: row of field
// col: column postion of field
void showEmptyField(unsigned char row, unsigned char col)
{
   leds[col*2] &= (~((unsigned int)0x0001 << (row*2)));
}


// show a hidden field (LED ON)
// row: row of field
// col: column postion of field
void showHiddenField(unsigned char row, unsigned char col)
{
   leds[col*2] |= ((unsigned int)0x0001 << (row*2));
}


// show an image
// row: row of field
// col: column postion of field
// *image: pointer to image array
void showImage(unsigned char row, unsigned char col, flash unsigned char *image)
{
   unsigned int value, mask;
   unsigned char i, j;
   
   for(i = 0; i < 5; i++)
   {
      value = *image++;
      mask = 0x7F;
      for(j = 0; j < row; j++)
      {
         value <<= 1;
         mask <<= 1;
      }
      leds[col] &= (~mask);
      leds[col] |= value;
      col++;
   }
}


// show which player may currently play 
// (6 LED pixels on left or right side in last display line)
void showCurrentPlayer(void)
{
   unsigned char i;
   
   if(cCurrentPlayer == 1)
   {                   
      for(i = 6; i < 12; i++)
      {
         leds[i] &= (~0x200);
      }
      for(i = 0; i < 6; i++)
      {
         leds[i] |= 0x200;
      }
   }
   else
   {
      for(i = 0; i < 6; i++)
      {
         leds[i] &= (~0x200);
      }
      for(i = 6; i < 12; i++)
      {
         leds[i] |= 0x200;
      }
   }
}



// show each players score
void showScore(void)
{
   unsigned char col, col2, i;
   
   clearScreen();
   
   showDecimal(cScorePlayer1, 2, 0);
   if(cScorePlayer1 >= 10)
   {          
      col = 6;
      col2 = 8;
   }                                
   else if(cScorePlayer2 >= 10)
   {          
      col = 4; 
      col2 = 6;
   }
   else                                
   {          
      col = 5; 
      col2 = 8;
   }
   leds[col] = 0x28;  // show colon :                                
   showDecimal(cScorePlayer2, 2, col2);
   if(cScorePlayer1 > cScorePlayer2)
   {
      for(i = 0; i < col; i++)
         ledsBlink[i] = 0xFFFF;
   }
   else
   {
      for(i = col+1; i < WIDTH; i++)
         ledsBlink[i] = 0xFFFF;
   }
}


// redraw all cards dependent on their states
void redrawCards(void)
{
   unsigned char cStatus;
   unsigned char row, col;

   for(row = 0; row < ROWS; row++)
   {
      for(col = 0; col < COLUMNS; col++)
      {
         cStatus = arCards[row][col].status;
         if(cStatus & CARDSELECTED)
         {
            selectField(row, col);
         }
         else
         {
            unselectField(row, col);
         }

         if( (cStatus & CARDPRESENT) == 0 )
         {  
            if( (cStatus & CARDSELECTED) == 0 )
            {
               showEmptyField(row, col);
            }
         }
         else
         {
            if(cStatus & CARDHIDDEN)
               showHiddenField(row, col);
         }
      }
   }
}


// show start logo
void showStartLogo(void)
{
   unsigned char i;
   flash unsigned int *pImage;

   pImage = logoMemo;
   for(i = 0; i < 12; i++)
   {
      leds[i] = *pImage++;
   }
}



// handle x/y-positions of potis: unselect previous field, select current field
void select(void)
{
   if( (cCurrentXpos != cXpos) || (cCurrentYpos != cYpos) )
   {
      if(cCurrentXpos != 0xFF)
      {
         arCards[cCurrentYpos][cCurrentXpos].status &= (~CARDSELECTED);
      }
      arCards[cYpos][cXpos].status |= CARDSELECTED;
      redrawCards();
   }
}



// handle keypress after a card was selected: show selected card, check for matching cards
void checkCards(void)
{
   static unsigned char cCardSelected = 0;
   unsigned char i;
   bit bCardsMatch = 0;

   if( (arCards[cYpos][cXpos].status & CARDPRESENT) == 0 )
      return;
   if(cCardSelected)
   {     // one card already shown
      if( (cXpos == cLastXpos) && (cYpos == cLastYpos) )
         return;
      arCards[cYpos][cXpos].status &= (~CARDHIDDEN);
      arCards[cYpos][cXpos].status &= (~CARDSELECTED);
      redrawCards();
  
      if(arCards[cYpos][cXpos].image == arCards[cLastYpos][cLastXpos].image)
      {   // cards match
         bCardsMatch = 1;
         arCards[cLastYpos][cLastXpos].status &= (~CARDPRESENT);
         arCards[cYpos][cXpos].status &= (~CARDPRESENT);
         redrawCards();
         if(cCurrentPlayer == 1)
            cScorePlayer1++;
         else
            cScorePlayer2++;

         cRemainingPairs--;
         if(cRemainingPairs == 0)
         {
            gameStatus = GAME_OVER;
         }
      }
      else
      {   // cards do not match
         arCards[cLastYpos][cLastXpos].status |= CARDHIDDEN;
         arCards[cYpos][cXpos].status |= CARDHIDDEN;
         redrawCards();
         cCurrentPlayer++;
         if(cCurrentPlayer > 2)
            cCurrentPlayer = 1;
      }
      cCardSelected = 0;
      redrawCards();
      
      clearScreen();
      showImage(1, 0, arCards[cLastYpos][cLastXpos].image);
      showImage(1, 7, arCards[cYpos][cXpos].image);
      if(bCardsMatch)
      {
         for(i = 0; i < WIDTH; i++)
            ledsBlink[i] = 0xFFFF;
      }
   }
   else
   {     // first card to show
      arCards[cYpos][cXpos].status &= (~CARDHIDDEN);
      arCards[cYpos][cXpos].status &= (~CARDSELECTED);
      redrawCards();
      clearScreen();
      showImage(1, 4, arCards[cYpos][cXpos].image);
      
      cCardSelected = 1;
   }
   cLastXpos = cXpos;
   cLastYpos = cYpos; 
   
   waitUntilKeyreleased();
   waitUntilKeypressed();
   arCards[cYpos][cXpos].status |= (CARDSELECTED);   
   clearScreen();
   if(bCardsMatch)
   {
      for(i = 0; i < WIDTH; i++)
         ledsBlink[i] = 0;
   }
   redrawCards();   
}


// setup variables, initializations
void setupGame (void)
{
  unsigned char row,col;
  unsigned char cnt = 0;
  unsigned char cImages;
  flash unsigned char *pImage = arImages;
  unsigned char i;
  unsigned char cValue;

  //initialize random generator and reset score
  srand(TCNT2);

  for(col = 0; col < COLUMNS; col++)
  {
     for(row = 0; row < ROWS; row++)
     {
        arCards[row][col].status = CARDPRESENT | CARDHIDDEN;
        arCards[row][col].image = 0;
     }
  }
  row = 0;
  col = 0;
  cnt = 0;
  for(cImages = 0; cImages < COLUMNS*ROWS; cImages++)
  {
     cValue = rand() % 100;
     cValue++;

     for(i = 0; i < cValue; i++)
     {
        do
        {
           if(row < MAX_ROW)
              row++;
           else
           {
              row = 0;
              if(col < MAX_COLUMN)
                 col++;
              else
                 col = 0;
           }
        }
        while(arCards[row][col].image != 0);
     }

     arCards[row][col].image = pImage;
     cnt++;
     if(cnt == 2)
     {
        cnt = 0;
        pImage += 5;
     }
  }
  
  clearScreen();
  cScorePlayer1 = 0;
  cScorePlayer2 = 0;
  cRemainingPairs = ROWS*COLUMNS/2;
  cXpos = 0;
  cYpos = 0;
  cCurrentXpos = 0xFF;
  redrawCards();

  showCurrentPlayer();
}



// show startup screen
void showStartupScreen(void)
{
   clearScreen();
   showStartLogo();
}


// control loop for memo game
void playMemo(void)
{
   unsigned char x, i;
   
   showStartupScreen();
   delay_ms(2000);
   clearScreen();
                
   do
   {
      x = selectX();
      if(x < 3)
      {
         cCurrentPlayer = 1;
         showLogo(arrowLeftLogo, BLINKING);
      }
      else
      {
         cCurrentPlayer = 2;
         showLogo(arrowRightLogo, BLINKING);
      }
   }
   while(!keyPressed());
   
   for(i = 0; i < WIDTH; i++)
   {
      ledsBlink[i] = 0;
   }
   setupGame();
   gameStatus = GAME_RUNNING;
   waitUntilKeyreleased();
   
   do
   {   
      cXpos = selectX();
      cYpos = selectY();
      select();
      cCurrentXpos = cXpos;
      cCurrentYpos = cYpos; 
      if(keyPressed())
      {
         checkCards();
         waitUntilKeyreleased();
         showCurrentPlayer();
      }
   }
   while(gameStatus != GAME_OVER);
    
   arCards[cYpos][cXpos].status &= (~CARDSELECTED);
   redrawCards();
   showScore();

   waitUntilKeypressed();
   waitUntilKeyreleased();
   for(i = 0; i < WIDTH; i++)
      ledsBlink[i] = 0;
}

